/*
 *      Main classes for Erik's Interest calculator.
 *
 *      This does pretty much everything.
 *
 *      Copyright (C) 2009 Erik Halvorson.
 *
 *      This file is part of Erik's Interest Calculator.
 *
 *      Erik's Interest Calculator is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by the Free Software
 *      Foundation, either version 3 of the License, or (at your option) any later version.
 *
 *      Erik's Interest Calculator is distributed in the hope that it will be useful, but
 *      WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *      FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 *      You should have recented a copy of the GNU General Public License along with Erik's
 *      Interest Calculator. If not, see <http://www.gnu.org/licenses/>.
 */

#include <QtGui>
#include "Interest.h"
#include <math.h>

int main(int argc, char **argv){
    QApplication a(argc, argv);
    Interest *i = new Interest();

	i->show();

	return a.exec();
}

Interest::Interest(QWidget *parent) : QMainWindow(parent){
    QGridLayout *lyt;
    QToolButton *clr;
	QWidget *cwidg;
	QRadioButton *bt;
    QMenuBar *mbar = new QMenuBar();
	QMenu *m;
	QAction *a;
	int i = 0;

	//  Create the menu bar..
    m = mbar->addMenu("File");

	a = m->addAction("New Window");
	connect(a, SIGNAL(triggered(bool)), this, SLOT(showNewWindow()));

	a = m->addAction("About");
	connect(a, SIGNAL(triggered(bool)), this, SLOT(showAbout()));

	a = m->addAction("Quit");
	connect(a, SIGNAL(triggered(bool)), this, SLOT(quit()));

    //  Construct the rest of the program.
    setWindowTitle("Erik's Interest Calculator");

	cwidg = new QWidget();
	lyt = new QGridLayout();

	valueEdit = new QLineEdit();
	defaultButtonGrp = new QButtonGroup();

    //  Create entries, etc for present value...
	i = 0;
	lyt->addWidget(new QLabel("Present Value:"), i, 0);
	lyt->addWidget(valueEdit, i, 1);
	clr = new QToolButton();
	clr->setText("X");
	lyt->addWidget(clr, i, 2);
	connect(clr, SIGNAL(clicked(bool)), valueEdit, SLOT(clear()));
	bt = new QRadioButton();
	lyt->addWidget(bt, i, 3);
	defaultButtonGrp->addButton(bt, 0);
	bt->click();
	i++;

	//  Interest rate...
	interestEdit = new QLineEdit();

	lyt->addWidget(new QLabel("Interest Rate: (APR %)"), i, 0);
	lyt->addWidget(interestEdit, i, 1);
	clr = new QToolButton();
	clr->setText("X");
	lyt->addWidget(clr, i, 2);
	connect(clr, SIGNAL(clicked(bool)), interestEdit, SLOT(clear()));
	bt = new QRadioButton();
	lyt->addWidget(bt, i, 3);
	defaultButtonGrp->addButton(bt, 1);
	i++;

    //  Future value...
	futureEdit = new QLineEdit();

	lyt->addWidget(new QLabel("Future Value:"), i, 0);
	lyt->addWidget(futureEdit, i, 1);
	clr = new QToolButton();
	clr->setText("X");
	lyt->addWidget(clr, i, 2);
	connect(clr, SIGNAL(clicked(bool)), futureEdit, SLOT(clear()));
	bt = new QRadioButton();
	lyt->addWidget(bt, i, 3);
	defaultButtonGrp->addButton(bt, 2);
	i++;

    //  Monthly payment value...
	paymentEdit = new QLineEdit();

	lyt->addWidget(new QLabel("Payment Value:"), i, 0);
	lyt->addWidget(paymentEdit, i, 1);
	clr = new QToolButton();
	clr->setText("X");
	lyt->addWidget(clr, i, 2);
	connect(clr, SIGNAL(clicked(bool)), paymentEdit, SLOT(clear()));
	bt = new QRadioButton();
	lyt->addWidget(bt, i, 3);
	defaultButtonGrp->addButton(bt, 3);
	i++;

    //  Number of payments...
	numberEdit = new QLineEdit();

	lyt->addWidget(new QLabel("Payment Count:"), i, 0);
	lyt->addWidget(numberEdit, i, 1);
	clr = new QToolButton();
	clr->setText("X");
	lyt->addWidget(clr, i, 2);
	connect(clr, SIGNAL(clicked(bool)), numberEdit, SLOT(clear()));
	bt = new QRadioButton();
	lyt->addWidget(bt, i, 3);
	defaultButtonGrp->addButton(bt, 4);
	i++;

    totalPayments = new QLabel();
	lyt->addWidget(totalPayments, i, 0, 1, 2);
	i++;

    interestPayments = new QLabel();
	lyt->addWidget(interestPayments, i, 0, 1, 2);
	i++;

    QPushButton *okButton;

    okButton = new QPushButton("Compute");
	connect(okButton, SIGNAL(clicked(bool)), this, SLOT(recompute()));
	lyt->addWidget(okButton, i, 1);

    cwidg->setLayout(lyt);
	this->setCentralWidget(cwidg);
}

void Interest::recompute(void){
    char empty = 0, which = 0;

	if(valueEdit->text().count() == 0){ empty++; which = 0; }
	if(interestEdit->text().count() == 0){ empty++; which = 1; }
	if(futureEdit->text().count() == 0){ empty++; which = 2; }
	if(paymentEdit->text().count() == 0){ empty++; which = 3; }
	if(numberEdit->text().count() == 0){ empty++; which = 4; }

    if(empty == 0){
        which = defaultButtonGrp->checkedId();
        //QMessageBox::critical(this, "Error", "You must leave a box empty to compute something.");
		//return;
    }
	else if(empty >= 2){
        QMessageBox::critical(this, "Error", "You can only compute one thing. Leave a box empty to compute something.");
		return;
    }
	else{
        QAbstractButton *bt = defaultButtonGrp->button(which);
		QRadioButton *rbt = qobject_cast<QRadioButton *>(bt);

		rbt->click();
	}

    //  Figure out what to recompute...
    switch(which){
        case 0: recomputePresentValue(); break;
		case 1: recomputeInterestValue(); break;
		case 2: recomputeFutureValue(); break;
		case 3: recomputePaymentValue(); break;
		case 4: recomputeNumberValue(); break;
    }
}

void Interest::recomputePresentValue(){
    double pres, inter, payment, num, future;
	double accum;
    double effrate;
    int i;

	inter = interestEdit->text().toDouble() / 100.0;
	future = futureEdit->text().toDouble();
	payment = paymentEdit->text().toDouble();
	num = numberEdit->text().toDouble();
	num = floor(num);

    effrate = inter / 12.0;
    accum = future;
	for(i = 0; i < num; i++){
        accum = accum / (1 + effrate) - payment;
    }

	pres = accum;

    totalPayments->setText(QString("Total Payments: %1").arg(payment * num));
    interestPayments->setText(QString("Interest: %1").arg(future - payment * num - pres));
  	valueEdit->setText(QString("%1").arg(accum));
}

void Interest::recomputeInterestValue(){
    double pres, payment, num, future;
	double accum;
    double effrate;

	pres = valueEdit->text().toDouble();
	future = futureEdit->text().toDouble();
	payment = paymentEdit->text().toDouble();
	num = numberEdit->text().toDouble();
	num = floor(num);

    double x, high, low;
	int nloop = 0;

    high = 1000.0;
	low = -1000.0;

	//  Now binary search.
	nloop = 0;
    do {
		x = (high + low) / 2.0;

        nloop++;
		if(nloop > 300){
            QMessageBox::critical(this, "Error", "The search loop has failed to converge. Try different input parameters.");
			return;
		}

	    effrate = x / (100 * 12.0);
		accum = evaluate(pres, effrate, num, payment);

		if(((pres < future)&&(future > accum))||((pres > future)&&(future < accum))){
            high = x;
        }
		else{
            low = x;
        }

    } while(fabs(accum - future) > 0.0005);

    totalPayments->setText(QString("Total Payments: %1").arg(payment * num));
    interestPayments->setText(QString("Interest: %1").arg(future - payment * num - pres));
  	interestEdit->setText(QString("%1").arg(x));
}

void Interest::recomputeFutureValue(){
    double pres, inter, payment, num;
	double accum;
    double effrate;

	pres = valueEdit->text().toDouble();
	inter = interestEdit->text().toDouble() / 100.0;
	payment = paymentEdit->text().toDouble();
	num = numberEdit->text().toDouble();
	num = floor(num);
	effrate = inter / (12);

	accum = evaluate(pres, effrate, num, payment);

    totalPayments->setText(QString("Total Payments: %1").arg(payment * num));
    interestPayments->setText(QString("Interest: %1").arg(accum - payment * num - pres));
  	futureEdit->setText(QString("%1").arg(accum));
}

void Interest::recomputePaymentValue(){
    double pres, inter, payment, num, future;
	double accum;
    double effrate;
    int i;

	pres = valueEdit->text().toDouble();
	inter = interestEdit->text().toDouble() / 100.0;
	future = futureEdit->text().toDouble();
	num = numberEdit->text().toDouble();
	num = floor(num);

	double high, low, x;

	effrate = inter / 12.0;

	if(pres < future){
		high = future - pres;
		low = 0.0;
		x = 1.0;
		do {

            accum = evaluate(pres, effrate, num, x);
	
			if(accum < future){
				x = x * 2.0;
			}
			else{
				high = x;
				x = x / 2.0;
				low = 0.0;
				break;
			}
		} while(true);
    }
	else{
        low = future - pres;
		high = 0.0;
		x = -1.0;
		do {
            accum = evaluate(pres, effrate, num, x);
	
			if(accum > future){
				x = x * 2.0;
			}
			else{
				low = x;
				x = x / 2.0;
				high = 0.0;
				break;
			}
		} while(true);
	}

	//  Binary search...
    do {
        accum = evaluate(pres, effrate, num, x);

		if(fabs(accum - future) < 0.0005) break;
		else if(accum < future){
            low = x;
			x = (high + low) / 2.0;
        }
		else{
            high = x;
			x = (high + low) / 2.0;
		}
    } while(fabs(future - accum) > 0.0005);

	accum = pres;
	payment = x;
	for(i = 0; i < num; i++){
		accum = accum * (1 + effrate) + x;
	}
	

    totalPayments->setText(QString("Total Payments: %1").arg(payment * num));
    interestPayments->setText(QString("Interest: %1").arg(accum - payment * num - pres));
  	paymentEdit->setText(QString("%1").arg(payment));

}

void Interest::recomputeNumberValue(){
    double pres, inter, payment, future;
	double accum;
    double effrate;
    int i;

	pres = valueEdit->text().toDouble();
	inter = interestEdit->text().toDouble() / 100.0;
	future = futureEdit->text().toDouble();
	payment = paymentEdit->text().toDouble();

    int tot = 0;
	double accum2;

	effrate = inter / 12.0;

    do {
		accum = pres;
		for(i = 0; i < tot; i++){
            accum2 = accum;
			accum = accum * (1 + effrate) + payment;
            if(fabs(accum - future) > fabs(accum2 - future)){
                QMessageBox::critical(this, "Error", "Payment is too small to reduce principal.");
				return;
            }
        }
		if(fabs(accum - future) < fabs(payment)){
            printf("%d + 1 payments will work...\n", tot);
			break;
		}
		else tot++;
    } while(true);

    accum2 = 0.0;
    accum = pres;
    for(i = 0; i < tot; i++){
        accum = accum * (1 + effrate) + payment;
		accum2 = accum2 + payment;
    }

	if(fabs(accum - future) > 0.0) accum2 = accum2 + future - accum;

	numberEdit->setText(QString("%1").arg(tot + 1));
    totalPayments->setText(QString("Total Payments: %1").arg(accum2));
    interestPayments->setText(QString("Interest: %1").arg(future - accum2 - pres));
}

double Interest::evaluate(double pres, double effrate, int num, double payment){
    double accum = pres;
    int i;

	for(i = 0; i < num; i++){
        accum = accum * (1 + effrate) + payment;
	}

	return accum;
}

void Interest::showAbout(void){
    QDialog *dlg;
	QGridLayout *gd;
	QPushButton *ok;
    QLabel *lbl;

	//  Create the window...
	dlg = new QDialog();
	dlg->setWindowTitle("About Erik's Interest Calculator");

    //  Layout, etc...
	gd = new QGridLayout();
	lbl = new QLabel(tr("Erik's Interest Calculator v0.5 beta<BR>Copyright (C) 2009 Erik Halvorson<BR>http://erikh.niahak.org/"));
	lbl->setMinimumSize(250, 150);
	lbl->setFrameStyle(QFrame::Box | QFrame::Raised);
	lbl->setLineWidth(1);
	lbl->setMidLineWidth(1);
	gd->addWidget(lbl, 0, 0, 1, 2);

	ok = new QPushButton("Ok");
	connect(ok, SIGNAL(clicked(bool)), dlg, SLOT(accept()));
	gd->addWidget(ok, 1, 1);

	ok = new QPushButton("License");
	gd->addWidget(ok, 2, 1);
	connect(ok, SIGNAL(clicked(bool)), this, SLOT(showLicense()));

	dlg->setLayout(gd);

	dlg->exec();
}

void Interest::showLicense(void){
    QDialog *dlg;
	QGridLayout *gd;
	QPushButton *ok;
    QLabel *lbl;
    QString str;

    dlg = new QDialog();
	dlg->setWindowTitle("Erik's Interest Calculator - License");

	gd = new QGridLayout();

	lbl = new QLabel();
	str = QString("");
	str.append("Erik's Interest Calculator - A simple financial planning tool\n");
	str.append("Copyright (C) 2009 Erik Halvorson\n\n");
	str.append("This program is free software: you can resdistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.");
	str.append("\n\n");
	str.append("This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.\n\n");
	str.append("You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.\n");
	lbl->setText(str);
	lbl->setWordWrap(true);
	lbl->setMinimumSize(350, 300);
	lbl->setFrameStyle(QFrame::Box | QFrame::Raised);
	lbl->setLineWidth(1);
	lbl->setMidLineWidth(1);

	gd->addWidget(lbl, 0, 0, 1, 2);

	ok = new QPushButton("Ok");
	connect(ok, SIGNAL(clicked(bool)), dlg, SLOT(accept()));
	gd->addWidget(ok, 1, 1);

	dlg->setLayout(gd);

	dlg->exec();
}

void Interest::showNewWindow(void){
    Interest *n;

	n = new Interest();

	n->show();
}

void Interest::quit(void){
    exit(0);
}
